/*

                Database Cache

    The database cache maintains a copy of the monthly log
    database on the desktop.  When items from the handheld
    are processed, the cache is responsible for determining
    if the record on the handheld has changed and, if so,
    updating the cache and any requested derivative files.

*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <assert.h>

/* #define TRACE */

#include "database.h"           
#include "cachedb.h"

struct dbMemCacheHeader **cacheT = NULL;  /* Cache pointer table */
static int cacheN = 0,            /* Number of items in cache */
       cacheL = 0,            /* Length of current cacheT */
       cacheDirty = 0;        /* Has cache been modified since opening */

#ifndef NDEBUG
/*  Encourage lots of expansions while debugging.  */
#define cacheInitial    2
#define cacheExpand 3
#else
#define cacheInitial    48        /* Initial cache table size */
#define cacheExpand 24        /* Cache expansion increment */
#endif

#ifndef NDEBUG
#define Trace(x)    fprintf(stderr, "%s\n", (x))
#else
#define Trace(x)
#endif

/*  dbCacheConsistent  --  Verify consistency of the cache.
               This does nothing if NDEBUG is defined.  */

#ifndef NDEBUG
static void dbCacheConsistent(void)
{
    int i;

    assert(cacheN <= cacheL);         /* Items in cache <= cache length */
    assert(cacheT != NULL);       /* Cache pointer table exists */

    for (i = 0; i < cacheN; i++) {
    assert(cacheT[i] != NULL);    /* No null pointers in cache slots */
    if (i > 0) {
        assert((cacheT[i - 1]->h.year < cacheT[i]->h.year) ||
           ((cacheT[i - 1]->h.year == cacheT[i]->h.year) &&
            (cacheT[i - 1]->h.month < cacheT[i]->h.month)));
    }
    }
}
#else
#define dbCacheConsistent()
#endif

/*  growCache  --  Check whether the cache is ready to accept
           an additional record and, if necessary,
           either create it from scratch or expand
           the existing cache to accommodate it.  */

static void growCache(void)
{
    if (cacheN >= cacheL) {
    if (cacheT == NULL) {
        cacheT = malloc(sizeof(struct dbMemCacheHeader *) * cacheInitial);
        assert(cacheT != NULL);
        memset(cacheT, 0, sizeof(struct dbMemCacheHeader *) * cacheInitial);
        cacheN = 0;
        cacheL = cacheInitial;
#ifdef TRACE
            fprintf(stderr, "cachedb: Cache created: initial size %d.\n", cacheL);
#endif
    } else {
        struct dbMemCacheHeader **nc;

        cacheL += cacheExpand;
        nc = malloc(sizeof(struct dbMemCacheHeader *) * cacheL);
        assert(nc != NULL);
        /*  Clear out new slots at the end of the cache.  */
        memset(nc + cacheN, 0, sizeof(struct dbMemCacheHeader *) * (cacheL - cacheN));
        memcpy(nc, cacheT, sizeof(struct dbMemCacheHeader *) * cacheN);
        free(cacheT);
        cacheT = nc;
#ifdef TRACE
            fprintf(stderr, "cachedb: Cache expanded by %d to %d records.\n", cacheExpand, cacheL);
#endif
    }
    }
    dbCacheConsistent();
}

/*  findInCache  --  Find the cache record, if present, for a
             given year and month.  This is only ever
             called, in practice, for months we know are
                     in the cache, so we don't bother to optimise
             for months not in the cache.  Still, it works
             and returns NULL if a month not in the cache is
             requested.  */

struct dbMemCacheHeader *findInCache(int year, int month)
{
    int i;

    assert(cacheT != NULL);
    for (i = 0; i < cacheN; i++) {
    if ((cacheT[i]->h.year == year) && (cacheT[i]->h.month == month)) {
        return cacheT[i];
    }
    }
    return NULL;
}

/*  dbCacheOpen  --  Open the cache for business, optionally
             loading it from a given file.  If NULL is
             passed as the file descriptor, a new
             memory cache is created.  */

void dbCacheOpen(FILE *fp)
{
    if (fp != NULL) {
    struct dbCacheHeader item;

    rewind(fp);           /* Just to be sure */
    while (fread(&item, sizeof item, 1, fp) == 1) {
        struct dbMemCacheHeader *m;
        int lr;

        item.flags &= ~dbchDirty; /* Clear record dirty flag */
        m = malloc((sizeof(struct dbMemCacheHeader) + item.rbytes) - 2);
        assert(m != NULL);
        memcpy(&(m->h), &item, sizeof item);
        lr = fread(m->dbrec, 1, m->h.rbytes, fp);
        assert(lr == m->h.rbytes);
        growCache();

            /*  Okay, now we know there's room in the cache to add the
        new item.  */

            assert(cacheN < cacheL);  /* We *do* know that, don't we ? */
        cacheT[cacheN++] = m;

        /*  The cache is supposed to be sorted in date order to
                permit binary searches.  When debugging, let's make
                sure that's the case after every record is added.  */

        dbCacheConsistent();
#ifdef TRACE
            fprintf(stderr, "cachedb: Loading record for %04d-%02d from file.\n",
        m->h.year, m->h.month + 1);
#endif
    }
    rewind(fp);           /* Just to be nice */
    }

    /*  The following code, which creates an empty cache in
    memory is not in an else clause of the test for a null
    file pointer because it is perfectly valid for a cache
    file to exist on disc but contain no records.  */

    growCache();
    cacheDirty = 0;           /* Mark no changes since open */
}

/*  dbCacheEmpty  --  Empty the cache, discarding all storage.  */

void dbCacheEmpty(void)
{
    if (cacheT != NULL) {
    int i;

    dbCacheConsistent();
    for (i = 0; i < cacheN; i++) {
        free(cacheT[i]);
    }
    free(cacheT);
    cacheT = NULL;
    cacheL = cacheN = 0;
    cacheDirty = 0;
    }
}

/*  dbCachePrint  --  Dump cache status on a given file.  */

void dbCachePrint(FILE *fo)
{
    if (cacheT == NULL) {
        fprintf(fo, "Cache is not open.\n");
    } else {
    int i;

    dbCacheConsistent();
        fprintf(fo, "Cache contains %d records.  Cache capacity %d records. Status: %s\n",
            cacheN, cacheL, cacheDirty ? "dirty" : "clean");
    for (i = 0; i < cacheN; i++) {
            fprintf(fo, "Record %3d: %04d-%02d %s",
        i, cacheT[i]->h.year, cacheT[i]->h.month,
        asctime(localtime(&(cacheT[i]->h.modtime))));
    }
    }
}

/*  dbCacheAdd  --  Add a record to the cache.  The return value
            is 1 if the cache was actually modified and
            zero if the record was already present in the
            cache.  */

int dbCacheAdd(int year, int month, unsigned char *rec, int recl)
{
    int modified = 0, i, j;
    struct dbMemCacheHeader *m;

    dbCacheConsistent();

    for (i = 0; i < cacheN; i++) {
    if ((year == cacheT[i]->h.year) &&
        (month ==  cacheT[i]->h.month)) {

        /* A record for this month is already present in the
           cache.  See if it needs to be updated. */

        if ((recl != cacheT[i]->h.rbytes) ||
        (memcmp(cacheT[i]->dbrec, rec, recl) != 0)) {
        if (recl != cacheT[i]->h.rbytes) {
#ifdef TRACE
                    fprintf(stderr, "cachedb: Resizing record for %04d-%02d at position %d from %d to %d bytes.",
            year, month, i, cacheT[i]->h.rbytes, recl);
#endif
            m = malloc((sizeof(struct dbMemCacheHeader) + recl) - 2);
            assert(m != NULL);
            memcpy(&(m->h), &(cacheT[i]->h), sizeof(struct dbMemCacheHeader));
            m->h.rbytes = recl;
            free(cacheT[i]);
            cacheT[i] = m;
        }
        assert(recl == cacheT[i]->h.rbytes);
        memcpy(cacheT[i]->dbrec, rec, recl);
        cacheT[i]->h.flags |= dbchDirty;  /* Mark record modified */
        time(&(cacheT[i]->h.modtime));
        cacheDirty = modified = 1;
        }
#ifdef TRACE
            fprintf(stderr, "cachedb: %s record for %04d-%02d at position %d in cache at %s",
                modified ? "Modified" : "Verified unchanged",
        year, month, i, asctime(localtime(&(cacheT[i]->h.modtime))));
#endif
        return modified;
    } else if ((year < cacheT[i]->h.year) ||
           ((year == cacheT[i]->h.year) && (month < cacheT[i]->h.month))) {
        cacheDirty = modified = 1;
        break;
    }
    }

    /*  At this point i is the record before which the new
    month should be added.  If (i == cacheN) the new month
    should be appended to the end of the cache.  */

    dbCacheConsistent();
    growCache();                      /* Make sure there's room for the new record */

    /*  Unless we're appending to the end of the cache, shift
    down any records for subsequent months to make room for
    this one.  */

    for (j = cacheN; j > i; j--) {
    cacheT[j] = cacheT[j - 1];
    }
    m = malloc((sizeof(struct dbMemCacheHeader) + recl) - 2);
    assert(m != NULL);
    m->h.year = year;
    m->h.month = month;
    m->h.rbytes = recl;
    m->h.flags = dbchDirty;       /* Record is new, and hence modified */
    time(&(m->h.modtime));
    memcpy(m->dbrec, rec, recl);
    cacheT[i] = m;
#ifdef TRACE
    fprintf(stderr, "cachedb: Added record for %04d-%02d at position %d in cache at %s",
    m->h.year, m->h.month, i, asctime(localtime(&(m->h.modtime))));
#endif
    cacheN++;

    dbCacheConsistent();
    return modified;
}

/*  dbIsCacheDirty  --  Determine whether the cache has been modified.  */

int dbIsCacheDirty(void)
{
    return cacheDirty;
}

/*  dbCacheWrite  --  Write cache to a binary file.  Note that this neither
              tests nor changes the cacheDirty flag.  */

void dbCacheWrite(FILE *fp)
{
    int i;

    assert(fp != NULL);

    dbCacheConsistent();
    for (i = 0; i < cacheN; i++) {
    fwrite(&(cacheT[i]->h), sizeof(struct dbCacheHeader), 1, fp);
    fwrite(cacheT[i]->dbrec, 1, cacheT[i]->h.rbytes, fp);
    }
}
