/*

            Monthly log HTML output

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "database.h"
#include "units.h"
#include "trendfit.h"
#include "chart.h"
#include "htmlout.h"

const char htmlPrefix[] = "\
<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 3.2 Final//EN\">\n\
<html version=\"-//W3C//DTD HTML 3.2 Final//EN\">\n\
<head>\n\
";

const char htmlTitle[] = "\
Eat Watch<sup><font face=\"Helvetica, Arial\" size=1>TM</font></sup>\
";

const char htmlSuffix[] = "\
<p>\n\
<hr>\n\
<a href=\"http://www.fourmilab.ch/hackdiet/\"><cite>The Hacker's Diet</cite></a><br>\n\
by <a href=\"http://www.fourmilab.ch/\">John Walker</a>\n\
";

const char * const weightUnits[3] = {
    "kilograms",
    "pounds",
    "stones"
};

const char * const weekdayslong[7] = {
                           "Sunday", "Monday", "Tuesday",
                           "Wednesday", "Thursday", "Friday", "Saturday"
                         },

           * const weekdays[7] = {
                           "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"
                         },

           * const months[12] =   {
                           "January", "February", "March",
                           "April", "May", "June",
                           "July", "August", "September",
                           "October", "November", "December"
                         };

extern const int joulesp;           /* Use kilojoules as enery unit ? */

/*  lastMonth  --  Determine month previous to this one.  */

static void lastMonth(int tyear, int tmon, int *pyear, int *pmon)
{
    if (tmon == 0) {
        *pyear = tyear - 1;
        *pmon = 11;
    } else {
        *pyear = tyear;
        *pmon = tmon - 1;
    }
}

/*  nextMonth  --  Determine month after this one.  */

static void nextMonth(int tyear, int tmon, int *nyear, int *nmon)
{
    if (tmon == 11) {
        *nyear = tyear + 1;
        *nmon = 0;
    } else {
        *nyear = tyear;
        *nmon = tmon + 1;
    }
}

/*  rungSpecified  --  Did any entry for this month specify an
                       exercise rung?  */

static int rungSpecified(struct hdMonth *m)
{
    int i;

    for (i = 0; i < ((int) m->days); i++) {
        if ((m->rungArray[i] != RUNG_ABSENT) &&
            (m->rungArray[i] > 0)) {
            return 1;
        }
    }
    return 0;
}

/*  htmlMonth  --  Output monthly log as HTML in a file name derived
                   from the month and year.  */

/*  Chart dimensions  */

#define ChartX      640
#define ChartY      480

/*  Legend dimensions  */

#define LegendNex   "lgndtwx.gif"     /* Rung present */
#define LegendXex   300
#define LegendYex    28

#define LegendN     "lgndtw.gif"      /* No rung in month */
#define LegendX     185
#define LegendY      28

void htmlMonth(struct hdMonth *m, time_t cachetime)
{
    int i, pyear, pmon, nyear, nmon, lday = -1, nflags = 0;
    char srung[4], swgt[6], strend[6], svar[6];
    char fn[24];
    FILE *fo;
    double tslope;
    signed long weeklyTrendChange, whole, frac;
    char caption[80], deltaw[12];
    int csgn, tsgn;
    long dcal;
    char *chartImage;
    Boolean mslope = false;
    struct tm today;                  /* Current date and time */
    time_t now;                       /* Current Unix second count */

    now = time(NULL);
    today = *localtime(&now);
    today.tm_year += 1900;
    today.tm_mon++;

    /*  Find last day of month for which weight is specified.  */

    for (i = m->days - 1; i >= 0; i--) {
        if (m->weightArray[i] != 0) {
            lday = i;
            break;
        }
    }

    /*  Determine number of flagged days in the month.  */

    for (i = 0; i < ((int) m->days); i++) {
        nflags += m->flagsArray[i] != 0;
    }

    chartImage = chartMonth(m, ChartX, ChartY, WHITE, -1);

    nextMonth(m->year, m->month, &nyear, &nmon);
    lastMonth(m->year, m->month, &pyear, &pmon);

    /*  If we're charting the current month, and the month is not
        complete, determine the slope for calculation of the
        weight loss/gain and calorie deficit/excess for the
        full month from the last day in this month's log to the
        corresponding day in last month's.  This gives a more
        accurate estimate of the trend than a short period entirely
        within the current month.  */
    
    if ((m->year == today.tm_year) && (m->month == (today.tm_mon - 1)) &&
        (((int) m->days) > today.tm_mday)) {
        int syear = m->year, smonth = m->month;
        
        if (lday >= 0) {
            if (smonth == 0) {
                smonth = 11;
                syear--;
            } else {
                smonth--;
            }
            mslope = fitSlopeToMultiMonthTrend(syear, smonth, lday,
                                               m->year, m->month, lday,
                                               &tslope);
        }
    }
    if ((!mslope) && (lday > 0)) {
        tslope = fitSlopeToTrend(m, 0, m->days - 1);
        mslope = true;
    }

    caption[0] = 0;

    if (mslope) {
        dcal = (long) (EnergyPerMonthlyWeightUnit(m, joulesp) * tslope);
        weeklyTrendChange = (signed long) (tslope * 1000000 * 7);
        weeklyTrendChange = cvtLong(weeklyTrendChange, (weightunit) m->weightUnit, (weightunit) m->weightUnit);

        tsgn = (weeklyTrendChange < 0);
        if (tsgn) {
            weeklyTrendChange = -weeklyTrendChange;
        }
        weeklyTrendChange += 5000L;
        whole = weeklyTrendChange  / 1000000L;
        frac = (weeklyTrendChange / 10000L) % 100;
        if (whole == 0 && frac == 0) {
            tsgn = 1;
        }
        csgn = (dcal <= 0);
        if (csgn) {
            dcal = -dcal;
        }
        sprintf(deltaw, "%ld.%02ld", whole, frac);
        if (strcmp(deltaw, "-0.00") == 0) {
            strcpy(deltaw, "0.00");
        }
        sprintf(caption, "Weekly %s %s %s.  Daily %s: %ld %s.",
                tsgn ? "loss" : "gain",
                deltaw,
                weightUnits[(m->weightUnit == wuStone) ? wuPound : m->weightUnit],
                csgn ? "deficit" : "excess",
                dcal,
        joulesp ? "kilojoules" : "calories");
    };
    if (nflags > 0) {
        sprintf(caption + strlen(caption), "%s%d%% flagged.",
            strlen(caption) > 0 ? "  " : "",
            (nflags * 100) / m->days);
    }

    /* Generate file name from ISO date. */

    sprintf(fn, "%04d-%02d.html", m->year, m->month + 1);
    fo = fopen(fn, "w");
    if (fo == NULL) {
        fprintf(stderr, "Cannot create output file %s\n", fn);
        exit(1);
    }

    fprintf(fo,"\
%s\
<!-- Cache-date: %lu -->\n\
<title>Eat Watch Log for %s %d</title>\n\
<meta name=\"author\" content=\"Eat Watch Report Generator\">\n\
</head>\n\
<body bgcolor=\"#FFFFFF\">\n\
<table width=\"100%%\">\n\
<tr>\n\
<td width=\"33%%\">&nbsp;\n\
<td width=\"33%%\" align=center>\n\
<table border=3 style=\"border-width=6px; border-color: #FF0000\" cellpadding=6>\n\
<tr bgcolor=\"#0000FF\">\n\
<th><font size=6 color=\"#FFFF00\" face=\"Helvetica, Arial\">%s %d</font>\n\
</table>\n\
<td width=\"33%%\" align=right>\n\
<a href=\"%04d-%02d.html\"><img src=\"previous.gif\" height=32 width=32 border=0 alt=\"%s %d\"></a>\n\
<a href=\"calendar.html#%d\"><img src=\"up.gif\" height=32 width=32 border=0 alt=\"Calendar\"></a>\n\
<a href=\"%04d-%02d.html\"><img src=\"next.gif\" height=32 width=32 border=0 alt=\"%s %d\"></a>\n\
</table>\n\
<center>\n\
<small><em>Weight in %s</em></small><br>\n\
",
        htmlPrefix,
        cachetime,
        months[m->month], m->year,
        months[m->month], m->year,
        pyear, pmon + 1, months[pmon], pyear,
        m->year,
        nyear, nmon + 1, months[nmon], nyear,
        weightUnits[m->weightUnit]
    );
    if (chartImage != NULL) {
        int rs = rungSpecified(m);

        fprintf(fo, "<img src=\"%s\" width=%d height=%d alt=\"Chart for %s %d\"><br>\n",
            chartImage, ChartX, ChartY, months[m->month], m->year);

        /*  Include appropriate legend depending on whether the month
            contains any exercise rung entries.  */

        fprintf(fo, "<img src=\"%s\" width=%d height=%d alt=\"Red line=trend, Blue/white diamond=daily weight%s\"><p>\n",
            rs ? LegendNex : LegendN,
            rs ? LegendXex : LegendX,
            rs ? LegendYex : LegendY,
            rs ? "Blue line=exercise rung" : "");
    }
    fprintf(fo,"\
<b>%s</b><p>\n\
<table width=\"75%%\" border cellpadding=2 bgcolor=\"#D0D0D0\">\n\
<tr>\n\
<th width=\"1%%\" colspan=2>Date\n\
<th width=\"1%%\">Weight\n\
<th width=\"1%%\">Trend\n\
<th width=\"1%%\">Var.\n\
<th width=\"1%%\">Rung\n\
<th width=\"1%%\">Flag\n\
<th>Comments\n\
",
        caption
    );

    for (i = 0; i < ((int) m->days); i++) {
        double var = (m->weightArray[i] / 100.0) - (m->trendArray[i] / 1000000.0);

        strcpy(srung, "&nbsp;");
        strcpy(swgt, "&nbsp;");
        strcpy(strend, "&nbsp;");
        strcpy(svar, "&nbsp;");

        if (m->rungArray[i] != RUNG_ABSENT) {
            sprintf(srung, "%d", m->rungArray[i]);
        }
        if (m->weightArray[i] != 0) {
            if (m->weightUnit == wuStone) {
                editWeight(swgt, m->weightArray[i], m->weightUnit);
                if (m->trendArray[i] != 0) {
                    editWeight(strend, (weight) TrendToWeight(m->trendArray[i]), m->weightUnit);
                }
            } else {
                sprintf(swgt, "%.1f", m->weightArray[i] / 100.0);
                if (m->trendArray[i] != 0) {
                    sprintf(strend, "%.1f", m->trendArray[i] / 1000000.0);
                }
            }
            sprintf(svar, "%+.1f", var);
            if ((strcmp(svar, "+0.0") == 0) || (strcmp(svar, "-0.0") == 0)) {
                strcpy(svar, "0.0");
            }
        }
        fprintf(fo, "<tr><td align=right>%d <td align=center>%s <td align=right>%s <td align=right>%s <td align=right><font color=\"#%s\">%s</font> <td align=right>%s <td align=center>%s <td>%s\n",
            i + 1, weekdays[(m->firstDay + i) % 7],
            swgt, strend,
            svar[0] == '-' ? "FF0000" : (svar[0] == '+' ? "0000FF" : "000000"), svar,
            srung,
            m->flagsArray[i] ? "X" : "&nbsp;",
            m->commentArray[i] == NULL ? "&nbsp;" : m->commentArray[i]);
    }

    fprintf(fo,"\
</table>\n\
</center>\n\
%s\n\
</body>\n\
</html>\n\
",
        htmlSuffix
    );
    fclose(fo);
}

/*  htmlCacheTime  --  Obtain cache time stored in an HTML file.  This
                       is used to determine if the database record from
                       which the HTML was generated has changed, necessitating
                       re-generation of the HTML.  If no HTML file exists for
                       the given year and month, a cache time of 0 is
                       returned, which forces generation.  */

time_t htmlCacheTime(int year, int month)
{
    char fn[24], s[132];
    char *cp;
    FILE *fi;
    time_t ctime = 0;                   /* Cache time */

    /* Generate file name from ISO date. */

    sprintf(fn, "%04d-%02d.html", year, month + 1);
    fi = fopen(fn, "r");
    if (fi != NULL) {
        int i;

        for (i = 0; i < 10; i++) {      /* Assume cache time saved in first 10 lines */
            if (fgets(s, sizeof s, fi) == NULL) {
                break;
            }
            if ((cp = strstr(s, "<!-- Cache-date: ")) != NULL) {
                ctime = atol(cp + 17);
                break;
            }
        }
        fclose(fi);
    }
    return ctime;
}
